#include <iostream>
#include <armadillo>
#include <cmath>
#include <boost/math/special_functions/gamma.hpp>
#include <boost/math/distributions/poisson.hpp>
#include "MJPsystems.h"
#include "MJPget_Q.h"
#include <chrono>

using namespace std;
// Flags for dealing with multiple time points - a single version is compiled
#define TT 1 // two tails (0 or 1 - but set to 2 to omit the code entirely)?
#define RN 1 // renormalise (0 or 1)

/*
void printdvec(double *pv, int n) {
  int i;
  for (i=0;i<n;i++) {
    cout << pv[i]<<", ";
  }
  cout <<"\n";
}
void printivec(int *pv, int n) {
  int i;
  for (i=0;i<n;i++) {
    cout << pv[i]<<", ";
  }
  cout <<"\n";
}
*/

inline double h(double x) {
  return 1.0-x+x*log(x);
}

inline double hifunc(double rho, double B,double loge) {
  return rho +(B-loge)/3.0*(1.0+sqrt(1.0+18.0*rho/(B-loge)));
}
inline double lofunc(double rho, double A, double loge) {
  const double logroot2pi=0.5*log(2*3.14159265);
  return rho+sqrt(2*rho)*sqrt(-logroot2pi-loge-1.5*log(A)+log(A-1));
}
unsigned int get_mlo(unsigned int mhi, double rho) {
  unsigned int mlo;
  // Since using unsigned int, need to be careful of negative numbers
  double dmlo=double(2*(int)(rho-0.5))-(double)mhi;
  if (dmlo>0) {
    mlo=(unsigned int) dmlo;
  }
  else {
    mlo=0;
  }
  // When using signed ints, was
  //  mlo=2*(int)(rho-0.5) - mhi;
  //  if (mlo<0) {
  //    mlo=0;
  //  }
  return mlo;
}
// Take the L1 norm of the matrix and return the mmax for the req precision
int get_m(double rho, double prec, unsigned int mlo=0) {
  const double logprec=log(prec), pi=3.14159265;
  double dmhi, dmlo;
  unsigned int mhi;
  
  dmhi= hifunc(rho,0.0,logprec)-1;
  dmlo=lofunc(rho,2*rho*h(dmhi/rho),logprec);
  if ((unsigned int)dmlo > mlo) {
    mlo=(unsigned int)dmlo;
  }
  
  if (log(boost::math::gamma_p((double)(mlo+1),rho))<logprec) {
    return mlo; // lower bound is the value we want
  }
  else {
    const double B=-0.5*log(4*pi*rho*h(dmlo/rho));
    if (B>logprec) {
      dmhi=hifunc(rho,B,logprec);
    }
    mhi=(unsigned int)(dmhi+1);
  
    //    cout<<mhi<<"-"<<mlo<<"=width: "<< mhi-mlo<<"\n";
  
    while (mhi-mlo>1) {
      unsigned int mmid=mlo+(mhi-mlo)/2; // =(mlo+mhi)/2, rounds down
      double dm=(double)mmid;
      double loginv;
      
      loginv=log(boost::math::gamma_p(dm+1,rho));
    //    cout <<mlo<<", "<<mmid <<", "<<mhi<<", "<<loginv<<"\n";

      if (loginv<logprec) {
	mhi=mmid;
      }
      else {
	mlo=mmid;
      }
    }
  }
  return mhi;
}

///////////////////////////////////////
/* Scaling and squaring - full matrix*/
///////////////////////////////////////
// Sparse Q
arma::mat SS_exp_Q(const arma::sp_mat Q, double prec,
		   bool renorm=true, bool t2=true, int *pm=NULL) {
  const int d=Q.n_cols;
  const double dd=(double)d;
  const double rho=-Q.min(), BIG=1e100;
  const unsigned int s= (rho>1) ? (int) (log(rho)/log(2)): 0;
  const double rhosmall=rho/exp((double)s*log(2));
  const arma::sp_mat Psmall=rhosmall*(Q/rho+speye(size(Q)));
  arma::mat mexpsmall;
  arma::mat prod=arma::mat(Psmall);
  const unsigned int mhi=get_m(rhosmall,prec/(1.0+(double)t2)),mlo=t2? get_mlo(mhi,rhosmall): 0;
  double fac=2.0, szprod, rncumu=0;
  unsigned int j;
  
  szprod=rhosmall;
  if (szprod>BIG) {
    prod/=szprod;
    szprod=1; rncumu=log(szprod);
  }
  if (mlo<=1) {
    mexpsmall=arma::mat(Psmall)+arma::eye(d,d); // 0th and 1st term
  }
  else {
    mexpsmall=arma::zeros(d,d);
  }

  //  cout << "mlo, s="<<mlo<<","<<s<<"\n";
  //  cout << Psmall <<"\n";
  
  j=2;
  while (j<=mhi) { // do vsum <- v^T e^P
    prod=prod*Psmall;
    prod/=fac;
    if (j>=mlo) {
      mexpsmall+=prod; 
    }
    szprod*=rhosmall/fac++; 
    if (szprod>BIG) {
      prod/=szprod; 
      if (j>=mlo) { 
	mexpsmall/=szprod; 
      }
      if (!renorm) {
	rncumu+=log(szprod);
      }
      szprod=1;
    }
    j++;
  }

  if (pm!=NULL) {
    *pm=mhi;
  }

  if (renorm) {
    mexpsmall/=(arma::accu(mexpsmall)/dd);
  }
  else {
    mexpsmall*=exp(rncumu-rhosmall);
  }

  //  cout << "mexpsmall\n"<<mexpsmall <<"\n";

  
  for (j=0; j<s;j++) { // repeated squaring
    mexpsmall *= mexpsmall;
  }
  //  cout << "mexp\n"<<mexpsmall <<"\n";

  return mexpsmall;
}
// Dense Q
arma::mat SS_exp_Q(const arma::dmat Q, double prec,
		   bool renorm=true, bool t2=true, int *pm=NULL) {
  const int d=Q.n_cols;
  const double dd=(double)d;
  const double rho=-Q.min(), BIG=1e100;
  const unsigned int s= (rho>1) ? (int) (log(rho)/log(2)): 0;
  const double rhosmall=rho/exp((double)s*log(2));
  const arma::dmat Psmall=rhosmall*(Q/rho+eye(size(Q)));
  arma::mat mexpsmall;
  arma::mat prod=arma::mat(Psmall);
  const unsigned int mhi=get_m(rhosmall,prec/(1.0+(double)t2)),mlo=t2? get_mlo(mhi,rhosmall): 0;
  double fac=2.0, szprod, rncumu=0;
  unsigned int j;
  
  szprod=rhosmall;
  if (szprod>BIG) {
    prod/=szprod;
    szprod=1; rncumu=log(szprod);
  }
  if (mlo<=1) {
    mexpsmall=arma::mat(Psmall)+arma::eye(d,d); // 0th and 1st term
  }
  else {
    mexpsmall=arma::zeros(d,d);
  }

  j=2;
  while (j<=mhi) { // do vsum <- v^T e^P
    prod=prod*Psmall;
    prod/=fac;
    if (j>=mlo) {
      mexpsmall+=prod; 
    }
    szprod*=rhosmall/fac++; 
    if (szprod>BIG) {
      prod/=szprod; 
      if (j>=mlo) { 
	mexpsmall/=szprod; 
      }
      if (!renorm) {
	rncumu+=log(szprod);
      }
      szprod=1;
    }
    j++;
  }

  if (pm!=NULL) {
    *pm=mhi;
  }

  if (renorm) {
    mexpsmall/=(arma::accu(mexpsmall)/dd);
  }
  else {
    mexpsmall*=exp(rncumu-rhosmall);
  }

  for (j=0; j<s;j++) { // repeated squaring
    mexpsmall *= mexpsmall;
  }

  return mexpsmall;
}
///////////////////////////////////////////
/* Scaling and squaring - vector * matrix*/
///////////////////////////////////////////
// Sparse version
arma::mat SS_v_exp_Q(const arma::mat v, const arma::sp_mat Q, double prec,
		     bool renorm=true, bool t2=true, int *pm=NULL) {
  const int d=v.n_cols;
  const double dd=(double)d;
  const double rho=-Q.min(), BIG=1e100;
  // For speed, minimise (approx) rho/2^s + s
  // Ignores potential gain in the scaled exp calculations from sparsity,
  // which is unreliable but would lead to larger rhosmall and smaller s.
  // Also ignores the O(sqrt(rhosmall)) additional terms needed for the
  // scaled exp, which would lead to smaller rhosmall & larger s.
  // And ignores the logd savings from replacing squaring with vec*mx ops.
  const unsigned int s= (rho>1) ? (unsigned int) (log(rho)/log(2)): 0;
  const unsigned int twops=(unsigned int)pow(2,s);
  const double rhosmall=rho/(double)twops;
  const arma::sp_mat Psmall=rhosmall*(Q/rho+speye(size(Q)));
  arma::mat mexpsmall;
  arma::mat prod=arma::mat(Psmall);
  const unsigned int mhi=get_m(rhosmall,prec/(1.0+(double)t2)),mlo=t2? get_mlo(mhi,rhosmall): 0;
  double fac=2.0, szprod, rncumu=0;
  arma::mat vtmp=v;
  unsigned int j;
  
  szprod=rhosmall;
  if (szprod>BIG) {
    prod/=szprod;
    szprod=1; rncumu=log(szprod);
  }
  if (mlo<=1) {
    mexpsmall=arma::mat(Psmall)+arma::eye(d,d); // 0th and 1st term
  }
  else {
    mexpsmall=arma::zeros(d,d);
  }

  //  cout << "mlo, s="<<mlo<<","<<s<<"\n";
  //  cout << Psmall <<"\n";
  
  j=2;
  while (j<=mhi) { // do vsum <- v^T e^P
    prod=prod*Psmall; // dense * sparse
    prod/=fac;
    if (j>=mlo) {
      mexpsmall+=prod; 
    }
    szprod*=rhosmall/fac++; 
    if (szprod>BIG) {
      prod/=szprod; 
      if (j>=mlo) { 
	mexpsmall/=szprod; 
      }
      if (!renorm) {
	rncumu+=log(szprod);
      }
      szprod=1;
    }
    j++;
  }

  if (pm!=NULL) {
    *pm=mhi;
  }

  if (renorm) {
    mexpsmall/=(arma::accu(mexpsmall)/(double)d);
  }
  else {
    mexpsmall*=exp(rncumu-rhosmall);
  }

  // Minimise s1 d^3 + 2^s2 d^2 subject to s1+s2=s.
  const unsigned int s2=(unsigned int)min((log(dd/log(2))/log(2)),(double)s); 
  const unsigned int s1=s-s2, np=(unsigned int)pow(2,s2);
  //    cout << "s, s1, s2, np: "<< s<< ", "<< s1 << ", "<< s2<<","<< np<<"\n";
  
  for (j=0; j<s1;j++) { // repeated squaring
    mexpsmall *= mexpsmall;
  }

  for (j=0;j<np;j++) { // vector-matrix product
    vtmp*=mexpsmall;
  }
  //  cout << "mexp\n"<<mexpsmall <<"\n";

  return vtmp;
}
// Dense version
arma::mat SS_v_exp_Q(const arma::mat v, const arma::mat Q, double prec,
		     bool renorm=true, bool t2=true, int *pm=NULL) {
  const int d=v.n_cols;
  const double dd=(double)d;
  const double rho=-Q.min(), BIG=1e100;
  // For speed, minimise (approx) rho/2^s + s
  // Ignores the O(sqrt(rhosmall)) additional terms needed for the
  // scaled exp, which would lead to smaller rhosmall & larger s.
  // And ignores the logd savings from replacing squaring with vec*mx ops.
  const unsigned int s= (rho>1) ? (unsigned int) (log(rho)/log(2)): 0;
  const unsigned int twops=(unsigned int)pow(2,s);
  const double rhosmall=rho/(double)twops;
  const arma::mat Psmall=rhosmall*(Q/rho+eye(size(Q)));
  arma::mat mexpsmall;
  arma::mat prod=arma::mat(Psmall);
  const unsigned int mhi=get_m(rhosmall,prec/(1.0+(double)t2)),mlo=t2? get_mlo(mhi,rhosmall): 0;
  double fac=2.0, szprod, rncumu=0;
  arma::mat vtmp=v;
  unsigned int j;

  szprod=rhosmall;
  if (szprod>BIG) {
    prod/=szprod;
    szprod=1; rncumu=log(szprod);
  }
  if (mlo<=1) {
    mexpsmall=arma::mat(Psmall)+arma::eye(d,d); // 0th and 1st term
  }
  else {
    mexpsmall=arma::zeros(d,d);
  }

  j=2;
  while (j<=mhi) { // do vsum <- v^T e^P
    prod=prod*Psmall; 
    prod/=fac;
    if (j>=mlo) {
      mexpsmall+=prod; 
    }
    szprod*=rhosmall/fac++; 
    if (szprod>BIG) {
      prod/=szprod; 
      if (j>=mlo) { 
	mexpsmall/=szprod; 
      }
      if (!renorm) {
	rncumu+=log(szprod);
      }
      szprod=1;
    }
    j++;
  }

  if (pm!=NULL) {
    *pm=mhi;
  }

  if (renorm) {
    mexpsmall/=(arma::accu(mexpsmall)/(double)d);
  }
  else {
    mexpsmall*=exp(rncumu-rhosmall);
  }

  // Minimise s1 d^3 + 2^s2 d^2 subject to s1+s2=s.
  const unsigned int s2=(unsigned int)min((log(dd/log(2))/log(2)),(double)s); 
  const unsigned int s1=s-s2, np=(unsigned int)pow(2,s2);
  //    cout << "s, s1, s2, np: "<< s<< ", "<< s1 << ", "<< s2<<","<< np<<"\n";
  
  for (j=0; j<s1;j++) { // repeated squaring
    mexpsmall *= mexpsmall;
  }

  for (j=0;j<np;j++) { // vector-matrix product
    vtmp*=mexpsmall;
  }
  //  cout << "mexp\n"<<mexpsmall <<"\n";

  return vtmp;
}

////////////////////
// Uniformisation //
////////////////////
// v is one or more row vectors
// Sparse Q
arma::mat Unif_v_exp_Q(const arma::mat v, const arma::sp_mat Q, double prec, bool renorm=true, bool t2=true, int *pm=NULL) {
  const double rho=-Q.min(), BIG=1e100;
  const int d=v.n_cols, nc=v.n_rows;
  const unsigned int mhi=get_m(rho,prec/(1.0+(double)t2)),mlo=t2? get_mlo(mhi,rho): 0;
  const double dnc=(double)nc;
  unsigned int j;
  const arma::sp_mat P=Q+rho*speye(size(Q));
  arma::mat vsum(nc,d),vpro=v;
  double fac=1.0, final_mult, szvpro, rncumu=0;

  //  cout << "rho="<<rho<<", mlo="<< mlo<< " ,mhi="<<mhi << ", d="<<d<<"\n";

  szvpro=arma::accu(v)/dnc;
  final_mult=szvpro;
  if (szvpro>BIG) {
    vpro/=szvpro;
    szvpro=1; rncumu=log(szvpro);
  }
  if (mlo==0) {
    vsum=vpro;
  }
  else {
    vsum.zeros();
  }

  j=1;
  while (j<=mhi) {

    vpro=vpro*P;
    vpro/=fac;
    if (j>=mlo) {
      vsum+=vpro; 
    }
    szvpro*=rho/fac++; 
    if ((szvpro>BIG)){
      vpro/=szvpro; 
      if (j>=mlo) { 
	vsum/=szvpro; 
      }
      if (!renorm) {
	rncumu+=log(szvpro);
      }
      szvpro=1;
    }
    j++;
  }

  if (pm!=NULL) {
    *pm=mhi;
  }
  if (renorm) {
    final_mult/=(arma::accu(vsum)/dnc);
  }
  else {
    final_mult=exp(rncumu-rho);
  }

  return vsum*final_mult;
}
// Dense Q
arma::mat Unif_v_exp_Q(const arma::mat v, const arma::mat Q, double prec, bool renorm=true, bool t2=true, int *pm=NULL) {
  const double rho=-Q.min(), BIG=1e100;
  const int d=v.n_cols, nc=v.n_rows;
  const unsigned int mhi=get_m(rho,prec/(1.0+(double)t2)),mlo=t2? get_mlo(mhi,rho): 0;
  const double dnc=(double)nc;
  unsigned int j;
  const arma::mat P=Q+rho*eye(size(Q));
  arma::mat vsum(nc,d),vpro=v;
  double fac=1.0, final_mult, szvpro, rncumu=0;
  
  //cout << "rho="<<rho<<", mlo="<< mlo<< " ,mhi="<<mhi << ", d="<<d<<"\n";

  szvpro=arma::accu(v)/dnc;
  final_mult=szvpro;
  if (szvpro>BIG) {
    vpro/=szvpro;
    szvpro=1; rncumu=log(szvpro);
  }
  if (mlo==0) {
    vsum=vpro;
  }
  else {
    vsum.zeros();
  }

  j=1;
  while (j<=mhi) {//&& (!cvgd)) { // do vsum <- v^T e^P
    vpro=vpro*P;
    vpro/=fac;
    if (j>=mlo) {
      vsum+=vpro; 
    }
    szvpro*=rho/fac++; 
    if ((szvpro>BIG)){
      vpro/=szvpro; 
      if (j>=mlo) { 
	vsum/=szvpro; 
      }
      if (!renorm) {
	rncumu+=log(szvpro);
      }
      szvpro=1;
    }
    j++;
  }

  if (pm!=NULL) {
    *pm=mhi;
  }
  if (renorm) {
    final_mult/=(arma::accu(vsum)/dnc);
  }
  else {
    final_mult=exp(rncumu-rho);
  }

  return vsum*final_mult;
}



/////////////////////////////////////////////////////////////////
// vector transpose * matrix exponential - choose which algorithm
// test for choice depends on input matrix type
/////////////////////////////////////////////////////////////////
// Sparse version
arma::mat v_exp_Q(const arma::mat v, const arma::sp_mat Q, double prec, bool renorm=true, bool t2=true, int *pm=NULL) {
  const int d=v.n_cols;
  double rho=-Q.min(), dd=(double)d;
  if (rho>4.0e9) {
    cout << "rho of "<<rho<<" set to 4.0e9\n";
    rho=4.0e9;
  }
  const int s= (rho>1) ? (int) (log(rho)/log(2)): 0;

  // Obvious formula for sparsity CPU cost, c = k_1 f d^2 doesnt hold well
  // and approximate k_1 >> k_2 where cost for non-sparse is k_2 d^2.
  // So approximate sparse cost as c = k_2 d^{1.5}
  if ((pow(dd,1.5)*(double)(1+s) < rho)) {
    if ((d>1000) && (rho>1000)) {
      cout << "SS " << d << ", "<< rho << "\n";
    }
    return SS_v_exp_Q(v, Q, prec, renorm, t2, pm);
  }
  else {
    if ((d>1000) && (rho>1000)) {
      cout << "Unif " << d << ", "<< rho << "\n";
    }
    return Unif_v_exp_Q(v, Q, prec, renorm, t2, pm);
  }
}
arma::mat vT_exp_Q(const arma::dvec &v, const arma::sp_mat &Q, double prec, bool renorm=true, bool t2=true, int *pm=NULL) {
  arma::dmat vT=v.t();
  return v_exp_Q(vT, Q, prec, renorm, t2, pm);
}
// Dense version
arma::mat v_exp_Q(const arma::mat v, const arma::mat Q, double prec, bool renorm=true, bool t2=true, int *pm=NULL) {
  const int d=v.n_cols;
  double rho=-Q.min(), dd=(double)d;
  if (rho>4.0e9) {
    cout << "rho of "<<rho<<" set to 4.0e9\n";
    rho=4.0e9;
  }
  const int s= (rho>1) ? (int) (log(rho)/log(2)): 0; // #squares in SS

  // Approximate k_1 >> k_2 where cost for non-sparse is k_2 d^2.
  if (dd*(double)(1+s) < rho) {
    if ((d>1000) && (rho>1000)) {
      cout << "SS " << d << ", "<< rho << "\n";
    }
    return SS_v_exp_Q(v, Q, prec, renorm, t2, pm);
  }
  else {
    if ((d>1000) && (rho>1000)) {
      cout << "Unif " << d << ", "<< rho << "\n";
    }
    return Unif_v_exp_Q(v, Q, prec, renorm, t2, pm);
  }
}
arma::mat vT_exp_Q(const arma::dvec &v, const arma::dmat &Q, double prec, bool renorm=true, bool t2=true, int *pm=NULL) {
  arma::dmat vT=v.t();
  return v_exp_Q(vT, Q, prec, renorm, t2, pm);
}
// Matrix action on a single vector but multiple time points supplied
// in a vector with t_j>=t_i for j>i.
#if TT == 0
#if RN == 0
arma::mat MUSPS_v_exp_spQ(arma::rowvec v, arma::sp_mat Q, double prec, arma::vec ts) {
  const double rho=-Q.min(), BIG=1e100;
  const int d=v.n_cols, nt=ts.n_elem, tmax=ts(nt-1);
  const arma::sp_mat P=Q+rho*speye(size(Q));
  int t,tlo=0;
  double fac=1.0;
  arma::mat vsum(d,nt); // output
  arma::rowvec vpro;
  unsigned int mmax, j;
  unsigned int *pmshi=new unsigned int[nt];
  double szvpro;
  double *prncu=new double[nt], *ptmultfac=new double[nt], *pts=ts.memptr();
  //  cout << "rho="<<rho<<", m="<< m<<"\n";

  // Since the ts are ordered, ms(j-1) is a lower bound for ms(j)
  pmshi[0]=get_m(rho*pts[0],prec);
  for (t=1;t<nt;t++) {
    pmshi[t]=get_m(rho*pts[t],prec,pmshi[t-1]);
  }
  //  printivec(pmshi,nt);
  mmax=pmshi[nt-1];

  szvpro=arma::sum(v);
  if (szvpro>BIG) {
    const double lsvp=log(szvpro);
    vpro=v/szvpro;
    fill(prncu,prncu+nt,lsvp);
    szvpro=1.0;
  }
  else {
    vpro=v;
    fill(prncu,prncu+nt,0.0);
  }
  fill(ptmultfac,ptmultfac+nt,1.0);
  
  for (t=0;t<nt;t++) {
    vsum.col(t)=vpro.t();
  }
  
  for (j=0;j<mmax;j++) { // do (e^P) vsum
    int tlotmp=tlo;
    vpro=(tmax/fac)*vpro*P;
    szvpro*=(rho*tmax/fac);

    if (szvpro>BIG) {
      const double lsvp=log(szvpro);
      for (t=tlo;t<nt;t++) {
	vsum.col(t)=vsum.col(t)/szvpro;
      }
      vpro=vpro/szvpro;
      for (t=tlo;t<nt;t++) {
      	prncu[t]=prncu[t]+lsvp;
      }
      szvpro=1.0;
    }

    for (t=tlo;t<nt;t++) {
      ptmultfac[t]=ptmultfac[t]*pts[t]/tmax; // stores (t/tmax)^{j+1}
    }
    for (t=tlo;t<nt;t++) {
      vsum.col(t)=vsum.col(t)+ptmultfac[t]*vpro.t();
      if (pmshi[t]==j-1) { // If have just done the last necessary P^m term
	tlotmp=t+1; // then don't bother with this one next time.
      }
    }
    tlo=tlotmp;
    fac++;
  }

  //  cout << "\ntlo end: "<< tlo<<"\n";
  
  for (t=0;t<nt;t++) {
    vsum.col(t)=vsum.col(t)*exp(prncu[t]-rho*pts[t]);
  }

  delete[] pmshi; delete[] prncu; delete[] ptmultfac;

  
  return vsum.t();

}
#else // TT=0, RN=1
arma::mat MUSPS_v_exp_spQ(arma::rowvec v, arma::sp_mat Q, double prec, arma::vec ts) {
  const double rho=-Q.min(), BIG=1e100;
  const int d=v.n_cols, nt=ts.n_elem, tmax=ts(nt-1);
  const arma::sp_mat P=Q+rho*speye(size(Q));
  int t,tlo=0;
  double fac=1.0;
  unsigned int mmax. j;
  arma::mat vsum(d,nt); // output
  arma::rowvec vpro, vcolsums;
  //  arma::vec ptmultfac=arma::ones(nt), prncu=arma::zeros(nt);
  unsigned int *pmshi=new unsigned int[nt];
  double szvpro, final_mult;
  double *ptmultfac=new double[nt], *pts=ts.memptr();
  //  cout << "rho="<<rho<<", m="<< m<<"\n";

  // Since the ts are ordered, ms(j-1) is a lower bound for ms(j)
  pmshi[0]=get_m(rho*pts[0],prec);
  for (t=1;t<nt;t++) {
    pmshi[t]=get_m(rho*pts[t],prec,pmshi[t-1]);
  }
  //  printivec(pmshi,nt);
  mmax=pmshi[nt-1];

  szvpro=arma::sum(v);
  final_mult=szvpro;
  if (szvpro>BIG) {
    vpro=v/szvpro;
    szvpro=1.0;
  }
  else {
    vpro=v;
  }
  fill(ptmultfac,ptmultfac+nt,1.0);
  
  for (t=0;t<nt;t++) {
    vsum.col(t)=vpro.t();
  }

  for (j=1;j<=mmax;j++) { // do (e^P) vsum
    int tlotmp=tlo;
    vpro=(tmax/fac)*vpro*P;
    szvpro*=(rho*tmax/fac);

    if (szvpro>BIG) {
      for (t=tlo;t<nt;t++) {
	vsum.col(t)=vsum.col(t)/szvpro;
      }
      vpro=vpro/szvpro;
      szvpro=1.0;
    }

    for (t=tlo;t<nt;t++) {
      ptmultfac[t]=ptmultfac[t]*pts[t]/tmax; // stores (t/tmax)^{j+1}
    }
    for (t=tlo;t<nt;t++) {
      vsum.col(t)=vsum.col(t)+ptmultfac[t]*vpro.t();
      if (pmshi[t]==j-1) { // If have just done the last necessary P^m term
	tlotmp=t+1; // then don't bother with this one next time.
      }
    }
    tlo=tlotmp;
    fac++;
  }


  vcolsums=arma::sum(vsum);
  for (t=0;t<nt;t++) {
    const double this_mult=final_mult/vcolsums(t);
    vsum.col(t)=vsum.col(t)*this_mult;
  }

  delete[] pmshi; delete[] ptmultfac;

  
  return vsum.t();

}
#endif
#elif TT == 1
#if RN == 0 // TT=1, RN=0
arma::mat MUSPS_v_exp_spQ(arma::rowvec v, arma::sp_mat Q, double prec, arma::vec ts) {
  const double rho=-Q.min(), BIG=1e100, SMALL=1e-100;
  const int d=v.n_cols, nt=ts.n_elem;
  double tmax=ts(nt-1);
  const arma::sp_mat P=Q+rho*speye(size(Q));
  int t,tlo,thi;
  double fac=1.0;
  unsigned int mmax,j;
  arma::mat vsum(d,nt); // output
  arma::rowvec vpro;
  unsigned int *pmshi=new unsigned int[nt], *pmslo=new unsigned int[nt+1];
  double szvpro;
  double *prncu=new double[nt], *ptmultfac=new double[nt], *pts=ts.memptr();
  //  cout << "rho="<<rho<<", m="<< m<<"\n";

  // Since the ts are ordered, ms(j-1) is a lower bound for ms(j)
  pmshi[0]=get_m(rho*pts[0],prec/2.0);
  pmslo[0]=get_mlo(pmshi[0],rho*pts[0]);
  for (t=1;t<nt;t++) {
    pmshi[t]=get_m(rho*pts[t],prec/2.0,pmshi[t-1]);
    pmslo[t]=get_mlo(pmshi[t],rho*pts[t]);
  }
  // NB: sequence of pmslo need not be monotonic, but this doesn't matter
  mmax=pmshi[nt-1];
  pmslo[nt]=mmax+1; // makes code logic easier later
  //  printivec(pmshi,nt);
  //  printivec(pmslo,nt+1);
  
  szvpro=arma::sum(v);
  if (szvpro>BIG) {
    const double lsvp=log(szvpro);
    vpro=v/szvpro;
    fill(prncu,prncu+nt,lsvp);
    szvpro=1.0;
  }
  else {
    vpro=v;
    fill(prncu,prncu+nt,0.0);
  }
  fill(ptmultfac,ptmultfac+nt,1.0);

  vsum.zeros();
  t=0;
  while (pmslo[t]==0) {
    vsum.col(t)=vpro.t();
    t++;
  }
  thi=t-1; tlo=0;

  j=1;
  while (j<=mmax) { // do (e^P) vsum
    while (pmslo[thi+1]<=j) {  // add any new times that have come into range
      thi++;          // need <= rather than == because pmslo not nec. monotone
    }
    while (pmshi[tlo]<j) { // remove any that have gone out of range
      tlo++;
    }
    //    cout << "j, tlo and thi="<<j<<", "<<tlo << ", "<<thi<<"\n";
    vpro*=P;
    vpro*=(tmax/fac); // tmax gives simpler (but less efficient) implementation
    szvpro*=(rho*tmax/fac);  // as it leads to more under/overflows

    if ((szvpro>BIG) || (ptmultfac[tlo]<SMALL)) {
      const double lsvp=log(szvpro);
      for (t=tlo;t<=thi;t++) {
	vsum.col(t)/=(ptmultfac[t]*szvpro);
      }
      for (t=tlo;t<nt;t++) {
      	prncu[t]=prncu[t]+lsvp+log(ptmultfac[t]);
      }
      fill(ptmultfac+tlo,ptmultfac+nt,1.0); 
      vpro/=szvpro;
      szvpro=1.0;
    }

    for (t=tlo;t<nt;t++) { // keep track for all that could be relevant
      ptmultfac[t]=ptmultfac[t]*pts[t]/tmax; 
    }
    for (t=tlo;t<=thi;t++) {
      vsum.col(t)+=ptmultfac[t]*vpro.t();
    }
    fac++; j++;
  }

  for (t=0;t<nt;t++) {
    vsum.col(t)*=exp(prncu[t]-rho*pts[t]);
  }

  delete[] pmslo; delete[] pmshi; delete[] prncu; delete[] ptmultfac;

  
  return vsum.t();

}
#else // TT=1, RN=1
arma::mat MUSPS_v_exp_spQ(arma::rowvec v, arma::sp_mat Q, double prec, arma::vec ts) {
  const double rho=-Q.min(), BIG=1e100, SMALL=1e-100;
  const int d=v.n_cols, nt=ts.n_elem;
  const arma::sp_mat P=Q+rho*speye(size(Q));
  int t,tlo,thi;
  unsigned int j;
  double fac=1.0;
  unsigned int mmax;
  arma::mat vsum(d,nt); // output
  arma::rowvec vpro, vcolsums;
  unsigned int *pmshi=new unsigned int[nt], *pmslo=new unsigned int[nt+1];
  double szvpro, final_mult;
  double *ptmultfac=new double[nt], *pts=ts.memptr();
  //  cout << "rho="<<rho<<", m="<< m<<"\n";
  
  // Since the ts are ordered, ms(j-1) is a lower bound for ms(j)
  pmshi[0]=get_m(rho*pts[0],prec/2.0);
  pmslo[0]=get_mlo(pmshi[0],rho*pts[0]);
  for (t=1;t<nt;t++) {
    pmshi[t]=get_m(rho*pts[t],prec/2.0,pmshi[t-1]);
    pmslo[t]=get_mlo(pmshi[t],rho*pts[t]);
  }
  //printivec(pmshi,nt);
  //printivec(pmslo,nt);
  mmax=pmshi[nt-1];
  pmslo[nt]=mmax+1; // makes code logic easier later
  
  szvpro=arma::sum(v);
  final_mult=szvpro;
  if (szvpro>BIG) {
    vpro=v/szvpro;
    szvpro=1.0;
  }
  else {
    vpro=v;
  }
  fill(ptmultfac,ptmultfac+nt,1.0);

  vsum.zeros();
  t=0;
  while (pmslo[t]==0) {
    vsum.col(t)=vpro.t();
    t++;
  }
  thi=t-1; tlo=0;

  j=1;
  while (j<=mmax) { // do (e^P) vsum
    while (pmslo[thi+1]<=j) {  // add any new times that have come into range
      thi++;
    }
    while (pmshi[tlo]<j) { // remove any that have gone out of range
      tlo++;
    }
    //    cout << "j, tlo and thi="<<j<<", "<<tlo << ", "<<thi<<"\n";

    vpro*=P;
    if (thi>-1) { // slightly more efficient (but slightly more complicated)
      const double mul=pts[thi]/fac;
      vpro*=mul;
      szvpro*=(rho*mul);
    }
    else {
      szvpro*=rho;
    }

    if ((szvpro>BIG) || (ptmultfac[tlo]<SMALL)) {
      for (t=tlo;t<=thi;t++) {
	vsum.col(t)/=(ptmultfac[t]*szvpro);
      }
      fill(ptmultfac+tlo,ptmultfac+thi+1,1.0);
      vpro/=szvpro;
      szvpro=1.0;
    }
    
    //    for (t=tlo;t<=thi;t++) {
    // }
    for (t=tlo;t<=thi;t++) {
      ptmultfac[t]=ptmultfac[t]*pts[t]/pts[thi];
      vsum.col(t)+=ptmultfac[t]*vpro.t();
    }
    fac++; j++;
  }

  vcolsums=arma::sum(vsum);
  for (t=0;t<nt;t++) {
    const double this_mult=final_mult/vcolsums(t);
    vsum.col(t)*=this_mult;
  }

  delete[] pmslo; delete[] pmshi; delete[] ptmultfac;

  
  return vsum.t();

}
#endif
#else // place holder function
arma::mat MUSPS_v_exp_spQ(arma::rowvec v, arma::sp_mat Q, double prec, arma::vec ts) {
  const int d=v.n_cols, nt=ts.n_elem;
  arma::mat vsum=arma::zeros(d,nt); // output
  return vsum.t();
}
#endif

arma::mat SPS_exp_spQ_v(arma::sp_mat Q, arma::mat v, double prec, bool ren=true, bool tt=true, int *pm=NULL) {
  //  arma::sp_mat Qt=Q.t();
  //arma::mat vt=v.t();
  return Unif_v_exp_Q(v.t(), Q.t(), prec, ren, tt, pm).t();
}
arma::mat MUSPS_exp_spQ_v(arma::sp_mat Q, arma::mat v, double prec, arma::vec ts) {
  return MUSPS_v_exp_spQ(v.t(), Q.t(), prec, ts).t();
}

/////////////////////////////
// Material for Krylov method
/////////////////////////////
// Dense
void Arnoldi(const int m, const arma::dvec &v, const arma::dmat &A, arma::dmat &V, arma::dmat &Hbar, double &beta) {
  arma::mat vnew;
  int i,j;

  Hbar.zeros();
  beta=arma::norm(v);
  vnew=v/beta;
  V.col(0)=vnew;

  for (j=0;j<m;j++) { // already have basis columns 0 to j
    vnew=A*vnew;
    for (i=0;i <= j; i++) {
      Hbar(i,j)=dot(vnew,V.col(i));
      vnew -= Hbar(i,j)*V.col(i);
    }
    Hbar(j+1,j)=arma::norm(vnew);
    vnew/=Hbar(j+1,j);
    V.col(j+1)=vnew;    
  }
}
// Sparse
void Arnoldi(const int m, const arma::dvec &v, const arma::sp_mat &A, arma::dmat &V, arma::dmat &Hbar, double &beta) {
  int i,j;
  beta=arma::norm(v);
  arma::dvec vnew=v/beta;
  V.col(0)=vnew;
  Hbar.zeros();
  
  for (j=0;j<m;j++) {
    vnew=A*vnew; // seems wrong way round for speed
    for (i=0;i <= j; i++) {
      Hbar(i,j)=dot(vnew,V.col(i));
      vnew -= Hbar(i,j)*V.col(i);
    }
    Hbar(j+1,j)=arma::norm(vnew);
    vnew/=Hbar(j+1,j);
    V.col(j+1)=vnew;
  }
}
// Arnoldi but with vector on the left - should be faster but isn't
void ArnoldiBw(const int m, const arma::dvec &v, const arma::sp_mat &A, arma::dmat &V, arma::dmat &Hbar, double &beta) {
  arma::dvec vnew;
  int i,j;

  Hbar.zeros();
  beta=arma::norm(v);
  vnew=v/beta;
  V.col(0)=vnew;

  for (j=0;j<m;j++) {
    vnew=(vnew.t()*A).t();
    for (i=0;i <= j; i++) {
      Hbar(i,j)=dot(vnew,V.col(i));
      vnew -= (Hbar(i,j)*V.col(i));
    }
    Hbar(j+1,j)=arma::norm(vnew);
    vnew/=Hbar(j+1,j);
    V.col(j+1)=vnew;
  }
}
// Dense
arma::dvec Krylov_exp_Q_v(const arma::dvec &v, const arma::dmat &Q) {
  const int d=(int)v.n_rows;
  const int m=d<30? d: 30;
  const double t=1.0;
  arma::dmat V(d,m+1);
  arma::dmat Hbar(m+1,m+1);
  double beta, tcumu=0.0;
  arma::dmat vlatest=v;

  while (tcumu<t) {
    Arnoldi(m,vlatest,Q,V,Hbar,beta);

    arma::cx_vec eignval=arma::eig_gen(Hbar);
    double mag=arma::max(arma::abs(eignval));
    //    double mag=arma::norm(Hbar,2);
    double tau=(double) m /(2*mag); // error approx 0.3033^m

    if (tau+tcumu>t) {
      tau=t-tcumu;
      tcumu=t+1e-10;
    }
    else {
      tcumu+=tau;
    }
    //    cout<<"tau and tcumu="<<tau<<" "<<tcumu<<"\n";
    arma::dmat exptauH=arma::expmat(tau*Hbar);
    vlatest=beta*V*exptauH.col(0);
  }
  return vlatest;
}
// Sparse
arma::dvec Krylov_exp_Q_v(const arma::dvec &v, const arma::sp_mat &Q) {
  const int d=(int)v.n_rows;
  const int m=d<30? d: 30;
  const double t=1.0;
  arma::dmat V(d,m+1);
  arma::dmat Hbar(m+1,m+1);
  double beta, tcumu=0.0;
  arma::dmat vlatest=v;
  
  while (tcumu<t) {
    Arnoldi(m,vlatest,Q,V,Hbar,beta); // takes 98% of CPU time
    arma::cx_vec eignval=arma::eig_gen(Hbar);
    double mag=arma::max(arma::abs(eignval));
    //    double mag=arma::norm(Hbar,2);
    double tau=(double) m /(2*mag); // error approx 0.3033^m = 3e-16 each time
    //    cout << "tau="<<tau<<"\n";
    
    if (tau+tcumu>t) {
      tau=t-tcumu;
      tcumu=t+1e-10;
    }
    else {
      tcumu+=tau;
    }

    arma::dmat exptauH=arma::expmat(tau*Hbar);
    vlatest=beta*V*exptauH.col(0);
  }

  return vlatest;
}
arma::dmat Krylov_vt_exp_Q(const arma::dvec &v, const arma::dmat &Q) {
  arma::dmat w;
  arma::dmat Qt=Q.t();
  w=Krylov_exp_Q_v(v,Qt);
  return w.t();
}

// Direct reverse - turns out to not be as efficient as the above!
arma::dmat Krylov_vt_exp_Qb(const arma::dvec &v, const arma::sp_mat &Q) {
  const int d=(int)v.n_rows;
  const int m=d<30? d: 30;
  const double t=1.0;
  arma::dmat V(d,m+1);
  arma::dmat Hbar(m+1,m+1);
  double beta, tcumu=0.0;
  arma::dvec vlatest=v;

  while (tcumu<t) {
    ArnoldiBw(m,vlatest,Q,V,Hbar,beta);
    
    //    double mag=arma::norm(Hbar,"2");
    arma::cx_vec eignval=arma::eig_gen(Hbar);
    double mag=arma::max(arma::abs(eignval));
    double tau=(double) m /(2*mag); // error approx 0.3033^m = 3e-16
    //    cout << "tau="<<tau<<"\n";
    
    if (tau+tcumu>t) {
      tau=t-tcumu;
      tcumu=t+1e-10;
    }
    else {
      tcumu+=tau;
    }
  
    arma::dmat exptauH=arma::expmat(tau*Hbar);
    vlatest=beta*V*exptauH.col(0);
  }

  return vlatest.t();
}

arma::dmat Krylov_vt_exp_Q(const arma::dvec &v, const arma::sp_mat &Q) {
  arma::dmat w;
  arma::sp_mat Qt=Q.t();
  w=Krylov_exp_Q_v(v,Qt);
  return w.t();
}


/*
int main(int argc, const char** pargv) {
  arma::vec xs;
  arma::vec LVthetas={.3,.4,.01};
  arma::sp_mat Q;
  arma::mat v;
  int i;
  int tst=0;

  if (argc>1) {
    tst=atoi(pargv[1]);
  }
  if (tst==-2) {
    double rho=262.0, eps=1e-5;
    unsigned int m1,m2,m3,m4;
    m1=get_m(rho,eps);
    m2=get_m(rho,eps,86);
    cout << m1 <<", "<<m2<<"\n";
    m1=get_m(1e7,1e-15);
    m2=get_m(1e8,1e-15);
    m3=get_m(1e9,1e-15);
    cout << m1 <<", "<<m2<<", "<<m3<<"\n";
    m4=get_m(4e9,1e-15);
    cout << m1 <<", "<<m2<<", "<<m3<<", "<<m4<<"\n";
    //  m4=get_m(1e10,1e-15); // fails: rho > largest unsigned int
  }
  if (tst==-1) { // understand creating matrices from pointers in armadillo
    double *pmy=new double[4];
    pmy[0]=0; pmy[1]=1; pmy[2]=2; pmy[3]=3;
    arma::mat my(pmy,2,2);
    cout <<my<<"\n";
    cout << arma::sum(my)<<"\n";
    delete[] pmy;
  }
  
  if (tst==0) { // check get_m against the system's qpois - values and timing
    double rho=0.1;
    double eps=1e-16;
    auto t1 = std::chrono::high_resolution_clock::now();
    double q1;
    int q3;
    int j;

    for (j=0;j<8;j++) {
      cout <<"rho="<<rho<<"\n";

    for (i=0;i<10000;i++) {
      boost::math::poisson dist(rho);
      q1=quantile(complement(dist,eps));
    }
    auto t2 = std::chrono::high_resolution_clock::now();
    cout<< "Timing for boost: "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count() <<"\n";

    for (i=0;i<10000;i++) {
      q3=get_m(rho,eps);
    }
    t1 = std::chrono::high_resolution_clock::now();
    cout<< "Timing for get_m: "<<std::chrono::duration_cast<std::chrono::milliseconds>(t1-t2).count() <<"\n";
    cout << q1 << ", "<< q3<<"\n";
    rho=rho*10.0;
    }
  }
  
  if (tst==1) { // test matrix exponential is producing the right answer
    arma::mat LoHi={{1,3},{3,4}};
    arma::mat v(7,1), vout;
    int i;
    double rho;
    for (i=0;i<6;i++) {
      v(i,0)=i+1;
    }
    v/=arma::accu(v);
    cout << v.t()<< "\n";
    Q=0.01*LoHi_create_Q(LoHi,LVthetas,LV_S,&LV_Rates);
    cout << arma::mat(Q) << "\n";
    rho=-Q.min();
    for (i=0;i<7;i++) {
      cout <<"i="<<i<<"\n";
      arma::mat Qdense=arma::mat(Q);
      unsigned int m=get_m(rho,1e-16);
      cout << "rho="<< rho << ", m=" << m << "\n";
      cout << SPS_exp_spQ_v(Q.t(),v,1e-16).t();// << "\n";
      cout << (arma::expmat(Qdense.t()) * v).t();// << "\n";
      cout << Unif_v_exp_Q(v.t(),Q,1e-16);// << "\n";
      cout << SS_v_exp_Q(v.t(),Q,1e-16);
      cout << SS_v_exp_Q(v.t(),Q,1e-16,false,false);
      cout << SS_v_exp_Q(v.t(),Q,1e-16,false,true);
      cout << SS_v_exp_Q(v.t(),Q,1e-16,true,false);
      cout << v_exp_Q(v.t(),Q,1e-16);
      cout << Unif_v_exp_Q(v.t(),Qdense,1e-16);
      cout << Unif_v_exp_Q(v.t(),Qdense,1e-16,false,false);
      cout << SS_v_exp_Q(v.t(),Qdense,1e-16);
      cout << SS_v_exp_Q(v.t(),Qdense,1e-16,false,false);
      cout << v_exp_Q(v.t(),Qdense,1e-16);
      cout << Krylov_vt_exp_Q(v,Qdense);
      Q*=10.0;
      rho*=10.0;
    }
  }
  if (tst==2) { // check code for multiple time points
    arma::mat LoHi={{1,3},{3,4}};
    arma::mat v(7,1), vout;
    int i;
    double eps=1e-9;
    for (i=0;i<7;i++) {
      v(i,0)=i+1;
    }
    v=v/28.0;
    Q=LoHi_create_Q(LoHi,LVthetas,LV_S,&LV_Rates);
    cout << Unif_v_exp_Q(v.t(),Q/10.0,eps) << "\n";
    cout << Unif_v_exp_Q(v.t(),Q/5.0,eps) << "\n";
    cout << Unif_v_exp_Q(v.t(),Q/2,eps) << "\n";
    cout << Unif_v_exp_Q(v.t(),Q*2.0,eps) << "\n";
    cout << Unif_v_exp_Q(v.t(),Q*2.1,eps) << "\n";
    cout << Unif_v_exp_Q(v.t(),Q*2.2,eps) << "\n";
    cout << Unif_v_exp_Q(v.t(),Q*20,eps) << "\n";
    cout << Unif_v_exp_Q(v.t(),Q*100.0,eps,true,true) << "\n";
    arma::vec ts={0.1,0.2,0.5,2.0,2.1,2.2,20,100.0};
    cout << "Times\n"<< ts.t() << "\n";
    cout << "MUSPS\n"<<MUSPS_v_exp_spQ(v.t(),Q,eps,ts) << "\n";
  }

  if (tst==3) { // timing with a big Q
    int npop=60, nts=100;
    arma::vec SEIRthetas={.01,.75,.25}; // R0=4
    arma::sp_mat Q=SEIR_create_Qfull(npop,SEIRthetas,SEIR_S,&SEIR_Rates)/100.0;
    int d=Q.n_cols;
    arma::mat v=arma::zeros(d,1);
    arma::mat v1,v2;
    arma::mat storage1=arma::mat(d,nts),storage2=arma::mat(d,nts);
    arma::mat storage3=arma::mat(d,nts),storage4=arma::mat(d,nts);
    arma::vec ts=arma::vec(nts);
    v(d-2)=1.0;
    double eps=1e-14;

    cout << "d="<<Q.n_rows <<", non-zero="<<Q.n_nonzero <<"\n";
    
    auto t1 = std::chrono::high_resolution_clock::now();
        for (i=0;i<nts;i++) {
      //      cout << i<<"\n";
          storage1.col(i)=Unif_v_exp_Q(v.t(),Q*(double)(i+1),eps/100).t();
        }
    auto t2 = std::chrono::high_resolution_clock::now();
    cout<< "Timing for naive: "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count() <<"\n";

    for (i=0;i<nts;i++) {
      ts(i)=(double)(i+1);
    }
    storage2=MUSPS_v_exp_spQ(v.t(),Q,eps,ts).t();
    t1 = std::chrono::high_resolution_clock::now();
    cout<< "Timing for Unif: "<<std::chrono::duration_cast<std::chrono::milliseconds>(t1-t2).count() <<"\n";

    cout<<"Discrepancy: "<<arma::sum(arma::sum(abs(storage2-storage1)))<< "\n";

    t1 = std::chrono::high_resolution_clock::now();
    storage3.col(0)=Unif_v_exp_Q(v.t(),Q,eps).t();
    for (i=1;i<nts;i++) {
      storage3.col(i)=Unif_v_exp_Q((storage3.col(i-1)).t(),Q,eps).t();
    }
    t2 = std::chrono::high_resolution_clock::now();
    cout<< "Timing for iterative: "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count() <<"\n";
    cout<<"Discrepancy: "<<arma::sum(arma::sum(abs(storage3-storage1)))<< "\n";

    t1 = std::chrono::high_resolution_clock::now();
    storage4.col(0)=Krylov_vt_exp_Q(v,Q).t();
    for (i=1;i<nts;i++) {
      storage4.col(i)=Krylov_vt_exp_Q((storage4.col(i-1)),Q).t();
    }
    t2 = std::chrono::high_resolution_clock::now();
    cout<< "Timing for iterative Krylov: "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count() <<"\n";
    cout<<"Discrepancy: "<<arma::sum(arma::sum(abs(storage4-storage1)))<< "\n";

  }

  if (tst==4) { // more thorough test of get_m
    int nrhos=7, neps=3;
    double prhos[]={.1,1,10,100,1000,10000,100000};
    double peps[]={1e-5,1e-9,1e-16};
    int nreps=1000,i,j,k;
    arma::mat pprob=arma::mat(neps,nrhos);
    arma::mat pms_CS=arma::mat(neps,nrhos);
    arma::mat pms_boost=arma::mat(neps,nrhos);

    if (argc>2) {
      nreps=atoi(pargv[2]);
      cout << "Nreps= "<< nreps << "\n";
    }
    
    for (i=0;i<neps;i++) {
      double eps=peps[i];
      for (j=0;j<nrhos;j++) {
	double rho=prhos[j];
	int m;
	auto t1 = std::chrono::high_resolution_clock::now();
	for (k=0;k<nreps;k++) {
	  m=get_m(rho,eps);
	}
	auto t2 = std::chrono::high_resolution_clock::now();
	auto tdiffgetm = std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count();
	pms_CS(i,j)=m-rho;

	t1 = std::chrono::high_resolution_clock::now();
	for (k=0;k<nreps;k++) {
	  boost::math::poisson dist(rho);
	  m=quantile(complement(dist,eps));
	}
	t2 = std::chrono::high_resolution_clock::now();
	auto tdiffboost=std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count();
	cout<< "Time with (rho,eps)="<<rho<<","<<eps<<"= [getm, boost] " << tdiffgetm<<", "<<tdiffboost<<"\n";
	pms_boost(i,j)=m-rho;
      }
    }
    cout << "ms_CS:\n"<< pms_CS <<"\n";
    //cout << "ms_boost:\n"<< pms_boost <<"\n";
    cout << "diff in ms:\n"<<pms_CS-pms_boost<<"\n";
    // so accuracy is same - NB R's qpois() gets the wrong answer for high rho and low epsilon
  }

  if (tst==5) { // timing with a big Q - single longish time
    int npop=100;
    arma::vec SEIRthetas={.01,.75,.25}; // R0=4
    arma::sp_mat Q=10.0*SEIR_create_Qfull(npop,SEIRthetas,SEIR_S,&SEIR_Rates);
    int d=Q.n_cols;
    arma::mat v=arma::zeros(d,1);
    v(d-2)=1.0;
    double eps=1e-14;

    cout << "d="<<Q.n_rows <<", non-zero="<<Q.n_nonzero <<"\n";
    
    auto t1 = std::chrono::high_resolution_clock::now();
    arma::vec Unifout=Unif_v_exp_Q(v.t(),Q,eps/100).t();
    auto t2 = std::chrono::high_resolution_clock::now();
    cout<< "Timing for unif: "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count() <<"\n";

    cout << "Krylov1\n";
    t1 = std::chrono::high_resolution_clock::now();
    arma::dvec Kryout=Krylov_vt_exp_Q(v,Q).t();
    t2 = std::chrono::high_resolution_clock::now();
    cout<< "Timing for Krylov: "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count() <<"\n";
    cout<<"Discrepancy: "<<arma::norm(Kryout-Unifout)<< "\n";

    t1 = std::chrono::high_resolution_clock::now();
    cout <<"Krylov2\n";
    arma::dvec Kryoutb=Krylov_vt_exp_Qb(v,Q).t();
    t2 = std::chrono::high_resolution_clock::now();
    cout<< "Timing for direct Krylov: "<<std::chrono::duration_cast<std::chrono::milliseconds>(t2-t1).count() <<"\n";
    cout<<"Discrepancy: "<<arma::norm(Kryoutb-Unifout)<< "\n";

  }

  
  return 0;
}

*/
